/*!
 * © 2022 David Vines
 *
 * domainOfTheAncients is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * any later version.
 *
 * domainOfTheAncients is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with domainOfTheAncients. If not, see https://www.gnu.org/licenses/gpl-2.0.html.
 */
/* globals ALLTECHS, TECHS, Map, ShipDesign, Ship */
/* exported Player */
class Player {
	constructor(stylename,capital,indexOfFavouredTech,baseSeed) {
		this._stylename = stylename;
		this._shipdesignidoffset = Player.OFFSET[stylename]; // Give ourselves an offset for ship design ids
		if (baseSeed) { this._seed = baseSeed + Player.OFFSET[stylename]; }
		this._ROTRNGState = "";
		this._indexOfFavouredTech = indexOfFavouredTech;
		this._tech = ALLTECHS[indexOfFavouredTech].option;
		this._carriedProduction = 0;
		this._research = Array.from(ALLTECHS, _tech => { return 0; });
		this._production = 400;
		this._shipDesigns = [];
		this._ships = [];
		this._hasBuiltAShip = false;
		this._capital = capital;
		this._map = null;
		this._active = true;
	}
	init(baseMap,turn) {
		this._map = new Map(baseMap);
		const stylename = this._stylename;
		const map = this._map;
		map.forEachHex(function(hex) {
			if (hex.owner == stylename) {
				map.setLastScanned(hex.id,turn,"");
			}
		});

	}
	getNextDesignId() {
		return this._shipdesignidoffset++;
	}
	get active() {
		return this._active;
	}
	deactivate() {
		this._active = false;
		this._ships = []; // Allow the ships to disappear (in effect on the next turn)
	}
	get indexOfFavouredTech() {
		return this._indexOfFavouredTech;
	}
	get map() {
		return this._map;
	}
	computeCommsGrid() {
		this._map.computeCommsGrid(this);
	}
	get stylename() {
		return this._stylename;
	}
	get capital() {
		return this._capital;
	}
	get production() {
		return this._production;
	}
	set production(p) {
		this._production = p;
	}
	get carriedProduction() {
		return this._carriedProduction;
	}
	set carriedProduction(cp) {
		this._carriedProduction = cp;
	}
	get techIndex() {
		return this._indexOfFavouredTech;
	}
	get tech() {
		return this._tech;
	}
	get troopValue() {
		const mults = [undefined, 1, 2, 5, 10, 20, 50, 100, 250];
		return mults[this.researchLevel(TECHS.planetaryDefences)];
	}
	get shipDesigns() {
		return this._shipDesigns;
	}
	get hasBuiltAShip() {
		return this._hasBuiltAShip;
	}
	addShipDesign(design) {
		this._shipDesigns.push(design);
	}
	getNumberOfDesigns() {
		return this._shipdesignidoffset - Player.OFFSET[this._stylename];
	}
	addShip(ship) {
		ship.report = "Construction completed at "+ship.location;
		this._ships.push(ship);
		this._hasBuiltAShip = true;
	}
	destroyShip(destroyedShip) {
		this._ships = this._ships.filter(x => x !== destroyedShip);
	}
	get ships() {
		return this._ships;
	}
	researchLevel(tech) {
		return ALLTECHS[tech.index].level(this._research[tech.index]);
	}
	addResearch(tech,points) {
		let oldLevel = this.researchLevel(tech);
		this._research[tech.index] += parseInt(points);
		let newLevel = this.researchLevel(tech);
		return newLevel > oldLevel;
	}
	forEachTech(f) {
		let self = this;
		ALLTECHS.forEach(function(tech,index) {
			f(tech,self._research[index]);
		});
	}
	get seed() {
		if (this._seed) {
			return this._seed;
		} else {
			return undefined;
		}
	}
	set seed(seed) {
		this._seed = seed;
	}
	get ROTRNGState() {
		return this._ROTRNGState;
	}
	set ROTRNGState(state) {
		this._ROTRNGState = state;
	}
	toJSON() {
		let data = {
			tech: this._indexOfFavouredTech,
			cf: this._carriedProduction,
			r: this._research,
			prod: this._production,
			bas: this._hasBuiltAShip,
			cap: this._capital,
			map: this._map,
			offset: this._shipdesignidoffset,
			designs: [],
			ships: [],
			ROT: this._ROTRNGState
		};
		if (!this._active) {
			data.dead = true;
		}
		if (this._seed) {
			data.seed = this._seed;
		}
		for(let design of this._shipDesigns) {
			data.designs.push(design);
		}
		for(let ship of this._ships) {
			data.ships.push(ship);
		}
		return data;
	}
	fromJSON(data,baseMap) {
		this._active = (data.dead ? false : true); // To allow active players to omit the field
		if (data.tech) { this._indexOfFavouredTech = data.tech; }
		this._production = data.prod;
		this._carriedProduction = data.cf;
		this._research = data.r;
		this._hasBuiltAShip = data.bas;
		if (data.seed) this._seed = data.seed;
		this._ROTRNGState = data.ROT;
		this._capital = data.cap;
		if (data.map) {
			this._map = new Map(data.map,baseMap);
		}
		this._shipdesignidoffset = data.offset;
		for(let design of data.designs) {
			this._shipDesigns.push(ShipDesign.fromJSON(design));
		}
		for(let ship of data.ships) {
			this._ships.push(Ship.fromJSON(ship));
		}
	}
}
Player.OFFSET = {'player': 2000, 'alkari': 100, 'auron':200, 'caffe':300, 'clanger':400, 'drahmini':500,
				'gobb':600, 'iyanden':700, 'jerandi':800, 'kalli':900, 'kelle':1000, 'pican':1100,
				'regul':1200, 'saggitt':1300, 'strata':1400, 'fleet':1500,'worati':1600
};